from typing import Optional, Union, Type

import dateutil.parser, json
from maltego_trx.maltego import MaltegoMsg, MaltegoTransform, MaltegoEntity
from stix2.base import _STIXBase
from stix2.exceptions import MissingPropertiesError, PropertyPresenceError, ObjectConfigurationError
from stix2.parsing import dict_to_stix2, _detect_spec_version, STIX2_OBJ_MAPS
from stix2.properties import TimestampProperty

from maltego_stix2.config import (
    _partial_reverse_type_map,
    _reverse_property_maps,
    _heritage_config,
)

def get_stix_type(stix_dict):
    v = _detect_spec_version(stix_dict)
    OBJ_MAP = dict(STIX2_OBJ_MAPS[v]['objects'], **STIX2_OBJ_MAPS[v]['observables'])
    return OBJ_MAP.get(stix_dict['type'])

normalizers = {
    TimestampProperty: dateutil.parser.parse
}


def convert_maltego_property_bag_to_stix2(maltego_properties, stix2_type, allow_custom_fields, assume_stix_input=False):
    reverse_property_map = _reverse_property_maps.get(stix2_type, {})
    # We assume no fields except the mapped ones will be valid in the spec
    res_dict = {}
    for k, v in maltego_properties.items():
        if not v or (k == "id" and not assume_stix_input):
            continue

        if not assume_stix_input and not (k in reverse_property_map or allow_custom_fields):
            continue

        mapped_stix_key = reverse_property_map.get(k, k)
        
        # Try to parse JSON dicts and lists
        if "[" in v or "{" in v:
            try:
                v = json.loads(v)
            except:
                pass

        res_dict[mapped_stix_key] = v

    if not assume_stix_input:
        res_dict.update({"type": stix2_type, "spec_version": "2.1"})

    return res_dict


def maltego_to_stix2(
    entity: MaltegoMsg, transform: MaltegoTransform=None, allow_custom_types=True, allow_custom_fields=True,
        allow_skipping_stix2_coercion=False
) -> Optional[Union[_STIXBase, dict]]:
    """
    Try to convert any incoming Maltego entity to the closest corresponding STIX2 entity.

    If the input is a "proper" STIX2 entity (i.e. it is in the "maltego.STIX2." namespace), the recovered entity will be
    (at least nearly) equivalent to the STIX2 JSON object that would have been used to generate the Maltego entity.

    If the input is any other Maltego entity, a best guess is applied to generate an equivalent STIX2 object. If no
    matching STIX2 type is found, None is returned unless allow_custom_types is set to True. If allow_custom_types is
    used, this function will generate arbitrary STIX2-like objects will be generated (e.g. a maltego.Person entity will
    result in a 'maltego-person' STIX2 object.

    Properties will be translated according to an internally defined mapping. For instance, in Maltego, the STIX field
    "value" of "domain-name" is instead called "fqdn" on the "STIX2.domain-name" Entity. This mapping will be undone by
    this function, to translate the "fqdn" field back to "value". If allow_custom_fields is True, *all* fields will be
    added to the output object,

    :param entity: an incoming MaltegoMsg object (see maltego-trx library)
    :param transform: reference to the transform object (pass in 'response' from the calling transform). If missing, no
        logs will be sent to the client.
    :param allow_custom_types: if True, arbitrary new types may be returned regardless of input type, if False (default)
        only return a result where an "official" STIX2 types can be found.
    :param allow_custom_fields: if True, *all* properties on the Maltego entitiy will be added to the output STIX
        object, no matter if a mapping exists for it or not. If False (default), only properties that can be explicitly
        mapped are kept on the output STIX object.
    :param allow_skipping_stix2_coercion: if True, the final dict is directly returned without being converted to a
        _STIXBase object first. By default, this is turned off.
    :return: The resulting _STIXBase object, or the equivalent dictionary if skip_stix2_coercion is True (or None).
    """
    is_proper_stix2 = False
    stix2_type = None
    for type_dict in entity.Genealogy:
        type_ = type_dict["Name"]
        if type_.startswith("maltego.STIX2."):
            is_proper_stix2 = True
            stix2_type = entity.Properties["type"]
            break
        elif type_ in _partial_reverse_type_map:
            stix2_type = _partial_reverse_type_map[type_]
            break
    if stix2_type is None:
        if allow_custom_types:
            stix2_type = (
                entity.Genealogy[0]["Name"].lower().replace(".", "-")
            )
        else:
            return None

    res_dict = convert_maltego_property_bag_to_stix2(
        entity.Properties, stix2_type, allow_custom_fields=allow_custom_fields, assume_stix_input=is_proper_stix2
    )

    obj_type: Type[_STIXBase] = get_stix_type(res_dict)
    if obj_type:
        for prop_name, prop_def in obj_type._properties.items():
            prop_value = res_dict.get(prop_name)
            if not prop_value:
                continue

            # TODO try to validate whether stix will accept the property.
            #  If not, try some recovery, if it fails, just delete the prop
            valid = False
            try:
                prop_def.clean(res_dict[prop_name])
                valid = True
            except:
                # TODO add more normalizers to recover
                normalizer = normalizers.get(prop_def.__class__)
                if normalizer:
                    try:
                        res_dict[prop_name] = normalizer(prop_value)
                        valid = True
                    except:
                        pass

            if not valid:
                if allow_custom_fields:
                    new_name = f"x_{prop_name}_unparseable"
                    if transform is not None:
                        transform.addUIMessage(
                            f"Warning: STIX2 conversion of property {prop_name} failed, it was added as {new_name} instead",
                            "PartialError"
                        )
                    res_dict[new_name] = prop_value
                else:
                    if transform is not None:
                        transform.addUIMessage(
                            f"Warning: STIX2 conversion of property {prop_name} failed and it was removed from the output",
                            "PartialError"
                        )
                    del res_dict[prop_name]
    elif "id" not in res_dict:
        res_dict["id"] = ""

    # Handle default values
    mapping = _heritage_config.get(stix2_type)
    if mapping is not None and mapping.default_values is not None:
        for k, v in mapping.default_values.items():
            if k not in res_dict:
                res_dict.update({k: v})

    failed = True
    reason = None
    extra_reasons = []
    stix2_object = res_dict
    try:
        stix2_object = dict_to_stix2(
            res_dict, allow_custom=allow_custom_types or allow_custom_fields
        )
        failed = False
    except MissingPropertiesError as e:
        reason = f"Missing properties error: {', '.join(e.args)}"
        for missing_prop_name in e.properties:
            if mapping is not None and missing_prop_name in mapping.property_map:
                maps_to = mapping.property_map[missing_prop_name]
                extra_reasons.append(
                    f"Note: the property '{missing_prop_name}' is likely called '{maps_to}' on your input Entity "
                    f"(and appears to be empty there)."
                )

    except PropertyPresenceError as e:
        reason = f"Property presence error ({e.__class__}): {', '.join(e.args)}"
    except ObjectConfigurationError as e:
        if hasattr(e, "reason"):
            reason = f"Object configuration error: {e.reason}"
        else:
            reason = f"Unknown object configuration error"
    except Exception as e:
        reason = f"Unknown exception occurred."

    if failed:
        if allow_skipping_stix2_coercion:
            if transform is not None:
                transform.addUIMessage(
                    f"Warning: Strict STIX2 conversion of object failed, "
                    f"object will be returned as-is and may not be fully STIX2 compliant. "
                    f"Reason of conversion failure: {reason}",
                    "PartialError"
                )
        else:
            if transform is not None:
                transform.addUIMessage(
                    f"Error: Strict STIX2 conversion of object failed, no output will be returned. "
                    f"Reason of failure: {reason}",
                    "PartialError"
                )
            stix2_object = None

    for r in extra_reasons:
        if transform is not None:
            transform.addUIMessage(r, "PartialError")

    return stix2_object


def stix2_to_maltego(stix2_object_or_dict: Union[_STIXBase, dict]) -> MaltegoEntity:
    """
    Given some STIX2 object (either as a _STIXBase object or equivalent dictionary), this function generates a new
    MaltegoEntity object that closely corresponds to the input STIX object.

    Objects, Properties and Value are automatically mapped from STIX to the closest corresponding internal Maltego
    equivalent to provide compatibility with other Maltego Transforms.

    :param stix2_object_or_dict: the _STIXBase object or dict to Translate
    :return: the created MaltegoEntity object
    """
    if not isinstance(stix2_object_or_dict, dict):
        stix2_object_as_dict: dict = stix2_object_or_dict._inner
    else:
        stix2_object_as_dict = stix2_object_or_dict

    stix_id = stix2_object_as_dict.get("id")
    if stix_id is None:
        raise ValueError("STIX2 objects must have an 'id' property.")

    mapping = _heritage_config.get(stix2_object_as_dict["type"])
    ent = MaltegoEntity(
        type=f"maltego.STIX2.{stix2_object_as_dict['type']}",
        value=stix_id  # value is always the STIX ID so that entities will merge properly
    )
    mtg_properties_set = set()
    for key, value in stix2_object_as_dict.items():
        if mapping is not None:
            key = mapping.translate_prop_name(key)
        if not key:
            continue
        if isinstance(value, dict) or isinstance(value, list):
            value = json.dumps(value)
        ent.addProperty(
            displayName=key, fieldName=key, value=value, matchingRule="loose"
        )
        mtg_properties_set.add(key)

    if mapping is not None and mapping.maltego_from_stix_property_map_extra is not None:
        for maltego_prop_name, stix_prop_name in mapping.maltego_from_stix_property_map_extra.items():
            if maltego_prop_name in mtg_properties_set:
                continue
            ent.addProperty(
                displayName=maltego_prop_name, fieldName=maltego_prop_name,
                value=stix2_object_as_dict[stix_prop_name],
                matchingRule="loose"
            )
    return ent




# Config of JSON-REF STIX 2 Schemas to process
# Each element should follow the format :
# {
#    "path": string, # path to a folder containing a JSON REF schema
#    "category": string   # Maltego Entity Category to group entities generated from this schema
# }
#
from collections import defaultdict

_schema_config = [
    {
        "path": "./cti-stix2-json-schemas/schemas/sdos/",
        "category": "STIX 2 domain objects",
        "tag": "default",
    },
    {
        "path": "./cti-stix2-json-schemas/schemas/sros/",
        "category": "STIX 2 relationship objects",
        "tag": "default",
    },
    {
        "path": "./cti-stix2-json-schemas/schemas/observables/",
        "category": "STIX 2 observables",
        "tag": "default",
    },
    {
        "path": "./cti-stix2-json-schemas-extended/schemas/sdos/",
        "category": "STIX 2 domain objects",
        "tag": "opencti",
    },  # These schemas allow for custom extensions of the STIX2 format. The provided ones are used in OpenCTI
    {
        "path": "./cti-stix2-json-schemas-extended/schemas/observables/",
        "category": "STIX 2 observables",
        "tag": "opencti",
    },  # These schemas allow for custom extensions of the STIX2 format. The provided ones are used in OpenCTI
]

digits = {
    "0": "zero", "1": "one", "2": "two", "3": "three", "4": "four", "5": "five", "6": "six", "7": "seven", "8": "eight",
    "9": "nine"
}

# Config of PNG files to process to create Maltego icons
# Each element should follow the format :
# {
#   "path": string, # path to a folder containing PNG files. A recursive search is done inside this folder.
#   "filter": string   # part of the PNG files names to process
#   "replace": {"string_to_replace": "replacement"} # Search and replace in PNG file names
# }
# The icon name should start with the type of the entity (eg "malware")
# If several icons are available for each entity, the filter parameter can be used to filter
# on the ones to keep. The icons names shoud contain this string just after the entity type.
_icons_config = [
    {
        "path": "./stix2-graphics/icons/png/",
        "filter": "-round-flat-300",
        "replace": {
            "coa": "course_of_action",
            "http": "url",
            "incident": "x_opencti_incident",
            **digits
        },
    },
    {
        "path": "./stix2-graphics-extended/icons/png/",
        "filter": "-round-flat-300",
        "replace": digits,
    },
]


class _MaltegoEntityMapping(object):
    def __init__(
        self,
        entity_type,
        property_map=None,
        display_value_override=None,
        icon_override=None,
        default_values=None,

        # this is for setting required maltego props that otherwise don't have a default (like URL's short-title)
        maltego_from_stix_extra_property_map=None,
        use_mapping_for_reverse_conversion=True,
    ):
        self.entity_type = entity_type
        self.property_map = property_map or {}
        self.icon_override = icon_override
        self.display_value_override = display_value_override
        self.default_values = default_values
        self.maltego_from_stix_property_map_extra = maltego_from_stix_extra_property_map
        self.use_mapping_for_reverse_conversion = use_mapping_for_reverse_conversion

    def translate_prop_name(self, stix_prop_name):
        return self.property_map.get(stix_prop_name, stix_prop_name)


_heritage_config = {
    # We add property mappings even for properties that happen to have the same name in order to take note of the fact
    # that these properties do in fact "make sense" within STIX2. This is important when we convert STIX2 Maltego
    # Entities back to STIX2 objects.
    "artifact": _MaltegoEntityMapping(None, None, None, icon_override="stix_two_default_icon"),  # was: Assemble
    "autonomous-system": _MaltegoEntityMapping(
        "maltego.AS",
        {"number": "as.number"},
        display_value_override="as.number",
    ),
    "domain-name": _MaltegoEntityMapping("maltego.Domain", {"value": "fqdn"}),
    "directory": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "file": _MaltegoEntityMapping(
        "maltego.File", {"path": "source", "name": "description"}, icon_override="stix_two_default_icon"  # was: Binary
    ),
    "email-addr": _MaltegoEntityMapping("maltego.EmailAddress", {"value": "email"}),
    "email-message": _MaltegoEntityMapping(
        "maltego.ConversationEmail",
        {"from_ref": "email", "to_refs": "email.recipients", "subject": "title"},
        icon_override="stix_two_email_msg",
        default_values={"is_multipart": False}
    ),
    "identity": _MaltegoEntityMapping("maltego.Organization", {"name": "title"}),
    "ipv4-addr": _MaltegoEntityMapping(
        "maltego.IPv4Address", {"value": "ipv4-address"}
    ),
    "location": _MaltegoEntityMapping(
        "maltego.Location",
        {
            "name": "location.name",
            "latitude": "latitude",
            "longitude": "longitude",
            "country": "country",
            "city": "city",
            "street_address": "streetaddress",
            "administrative_area": "location.area",
            "postal_code": "location.areacode",
        },
    ),
    "mac-addr": _MaltegoEntityMapping(
        "maltego.MacAddress", {"value": "macaddress"}
    ),  # added
    "threat-actor": _MaltegoEntityMapping(
        "maltego.Organization", {"name": "title"},
        use_mapping_for_reverse_conversion=False  # "every threat actor is an org but not vice versa"
    ),
    "url": _MaltegoEntityMapping(
        "maltego.URL", {"value": "url"}, maltego_from_stix_extra_property_map={"short-title": "value"}
    ),
    "user-account": _MaltegoEntityMapping(
        "maltego.Alias", {"user_id": "alias"}, display_value_override="alias"
    ),
    "x509-certificate": _MaltegoEntityMapping(
        "maltego.X509Certificate",
        {
            "subject": "subject",
            "serial_number": "serial",
            "issuer": "issuer",
            "validity_not_before": "validFrom",
            "validity_not_after": "validTo",
        },
        icon_override="stix_two_default_icon",
    ),
    #"note": None,
    "ipv6-addr": _MaltegoEntityMapping(
        "maltego.IPv6Address", {"value": "ipv6-address"}
    ),
    "x-opencti-hostname": _MaltegoEntityMapping("maltego.DNSName", {"value": "fqdn"}, icon_override="stix_two_default_icon"),
    "x-opencti-incident": _MaltegoEntityMapping("maltego.Incident", {"name": "title"}),
    "campaign": _MaltegoEntityMapping(
        "maltego.Event",
        {
            "name": "title",
            "first_seen": "starttime",
            "last_seen": "stoptime"
        }
    ),
    # "course-of-action": None,
    # "grouping": None,
    # "report": None,
    "vulnerability": _MaltegoEntityMapping(
        # Not 100% clean inheritance, since some Vulnerabilities are not CVE's.
        # However, maltego.CVE inherits maltego.Phrase, so for search Transforms this ends up working out fine
        # most of the time. Since we cannot retroactively make CVE inherit from Vulnerability, this is the next best
        # option for the time being.
        "maltego.CVE",
        property_map={"name": "text"}
        # icon_override="stix_two_default_icon"
    ),
    # "malware": None,
    # "observed-data": None,
    # "tool": None,
    # "attack-pattern": None,
    # "malware-analysis": None,
    # "opinion": None,
    # "indicator": None,
    # "intrusion-set": None,
    # "infrastructure": None,
    # "relationship": None,
    # "sighting": None,
    # "network-traffic": None,
    "windows-registry-key": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "process": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "software": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "mutex": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "x-opencti-text": _MaltegoEntityMapping(
        "maltego.Phrase",
        {"value": "text"},
        display_value_override="text",
        icon_override="stix_two_default_icon"
    ),
    "x-opencti-cryptographic-key": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "x-opencti-cryptocurrency-wallet": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon"),
    "x-opencti-user-agent": _MaltegoEntityMapping(None, icon_override="stix_two_default_icon")
}

# can be used to initialize "best-guess" STIX2 from arbitrary Maltego entities
# TODO maybe it makes more sense to hard-code an extended version of this so we can cover more Maltego entities
_partial_reverse_type_map = {
    v.entity_type: k for k, v in _heritage_config.items()
    if v.entity_type is not None and v.use_mapping_for_reverse_conversion
}

# can be used to convert STIX2 Maltego entities back into proper STIX2 entities
_reverse_property_maps = defaultdict(dict)
for class_name, mapping in _heritage_config.items():
    if mapping.entity_type is not None:
        for _k, _v in mapping.property_map.items():
            if _v in _reverse_property_maps[class_name]:
                raise ValueError(
                    f"Invalid property mapping for {class_name} ({_v} is already mapped)."
                    " Mapping a stix property to multiple Maltego properties break conversion and is not allowed."
                    " If you need to always populate a required field, please use the 'default_values' parameter."
                )
            _reverse_property_maps[class_name][_v] = _k





Artifact = "maltego.STIX2.artifact"
AttackPattern = "maltego.STIX2.attack-pattern"
AutonomousSystem = "maltego.STIX2.autonomous-system"
Campaign = "maltego.STIX2.campaign"
CourseOfAction = "maltego.STIX2.course-of-action"
Directory = "maltego.STIX2.directory"
DomainName = "maltego.STIX2.domain-name"
EmailAddr = "maltego.STIX2.email-addr"
EmailMessage = "maltego.STIX2.email-message"
File = "maltego.STIX2.file"
Grouping = "maltego.STIX2.grouping"
Identity = "maltego.STIX2.identity"
Indicator = "maltego.STIX2.indicator"
Infrastructure = "maltego.STIX2.infrastructure"
IntrusionSet = "maltego.STIX2.intrusion-set"
Ipv4Addr = "maltego.STIX2.ipv4-addr"
Ipv6Addr = "maltego.STIX2.ipv6-addr"
Location = "maltego.STIX2.location"
MacAddr = "maltego.STIX2.mac-addr"
Malware = "maltego.STIX2.malware"
MalwareAnalysis = "maltego.STIX2.malware-analysis"
Mutex = "maltego.STIX2.mutex"
NetworkTraffic = "maltego.STIX2.network-traffic"
Note = "maltego.STIX2.note"
ObservedData = "maltego.STIX2.observed-data"
Opinion = "maltego.STIX2.opinion"
Process = "maltego.STIX2.process"
Relationship = "maltego.STIX2.relationship"
Report = "maltego.STIX2.report"
Sighting = "maltego.STIX2.sighting"
Software = "maltego.STIX2.software"
ThreatActor = "maltego.STIX2.threat-actor"
Tool = "maltego.STIX2.tool"
Url = "maltego.STIX2.url"
UserAccount = "maltego.STIX2.user-account"
Vulnerability = "maltego.STIX2.vulnerability"
WindowsRegistryKey = "maltego.STIX2.windows-registry-key"
XOpenctiCryptocurrencyWallet = "maltego.STIX2.x-opencti-cryptocurrency-wallet"
XOpenctiCryptographicKey = "maltego.STIX2.x-opencti-cryptographic-key"
XOpenctiHostname = "maltego.STIX2.x-opencti-hostname"
XOpenctiIncident = "maltego.STIX2.x-opencti-incident"
XOpenctiText = "maltego.STIX2.x-opencti-text"
XOpenctiUserAgent = "maltego.STIX2.x-opencti-user-agent"
X509Certificate = "maltego.STIX2.x509-certificate"



from setuptools import setup
from maltego_stix2 import VERSION

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

dependencies = [
    'Pillow==8.1.1',
    'maltego-trx==1.3.8',
    'stix2==2.1.0',
    'python-dateutil==2.8.1'
]

setup(
    name='maltego-stix2',
    version=VERSION,
    description='Helper utilities for building Maltego transforms that consume or produce data in STIX2 format.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/amr-cossi/maltego-stix2',
    author='ANSSI, Samuel Hassine, Maltego Technologies GmbH',
    author_email='moa.amr.cossi@ssi.gouv.fr, samuel.hassine@luatix.org, support@maltego.com',
    license='Apache License 2.0',
    setup_requires=dependencies,
    install_requires=dependencies,
    packages=[
        'maltego_stix2'
    ],
    zip_safe=False
)
